#!/usr/bin/env python

# vim: tabstop=4 shiftwidth=4 softtabstop=4
#
#  Copyright (c) 2011 Openstack, LLC.
#  All Rights Reserved.
#
#     Licensed under the Apache License, Version 2.0 (the "License"); you may
#     not use this file except in compliance with the License. You may obtain
#     a copy of the License at
#
#          http://www.apache.org/licenses/LICENSE-2.0
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#     WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#     License for the specific language governing permissions and limitations
#     under the License.
#

import os
import re
import shutil
import subprocess
import sys


def install_libs(binary, installdir):
    """
    Install all dynamic library dependencies for a binary
    """
    # Strip extra leading slashses
    while installdir.startswith('//'):
        installdir = installdir[1:]

    def _find_libs(target):
        """
        Use ldd on a binary/library to find out its dynamic libraries.
        """

        p = subprocess.Popen(["ldd", target],
                stdin=subprocess.PIPE,
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE)
        (outdata, errdata) = p.communicate()

        libs = set()

        for line in outdata.split('\n'):
            fields = line.split()

            if not len(fields):
                continue

            if len(fields) > 2 and os.path.exists(fields[2]):
                    libs.add(fields[2])
            elif os.path.exists(fields[0]):
                    libs.add(fields[0])

        return libs

    def find_libs(target):
        """
        Get a list of libraries for a target.  Recurse through
        those libraries to find other libraries.
        """

        libs = set()
        more_libs = _find_libs(target)

        while libs != more_libs:
            for lib in set(more_libs - libs):
                libs.add(lib)
                more_libs.update(_find_libs(lib))

        return libs

    for lib in find_libs(binary):
        fname = os.path.basename(lib)
        if os.path.exists(installdir + '/' + fname):
            # Already installed
            continue
        print("Installing %s" % lib)
        shutil.copy2(lib, installdir)


def fix_ubuntu12_lib(data_dir):
    """ Ubuntu 12.04 Fix for include and config resources. """
    def copy_resource(src, dest):
        if not os.path.exists(dest) and os.path.exists(src):
            shutil.copytree(src, dest)

    import platform
    if set(["Ubuntu", '12.04']).issubset(set(platform.dist())):
        major, minor = sys.version_info[0:2]
        python_ver = "python%s.%s" % (major, minor)
        for path in sys.path:
            if os.path.exists("%s/config" % path):
                config_source = os.path.join(path, "config")
        config_destination = os.path.join(data_dir,
                                          "lib",
                                          python_ver,
                                          "config")
        include_source = os.path.join("/usr/include", python_ver)
        include_destination = os.path.join(data_dir,
                                           "local",
                                           "include",
                                          python_ver)

        copy_resource(config_source, config_destination)
        copy_resource(include_source, include_destination)


if __name__ == "__main__":
    if len(sys.argv) != 4:
        print "Usage: install_libs.py <binary_name> <data_dir> <install_dir>"
        sys.exit(1)

    binary = sys.argv[1]
    datadir = sys.argv[2]
    installdir = sys.argv[3]

    if not os.path.exists(installdir):
        os.makedirs(installdir)
    elif not os.path.isdir(installdir):
        print "Error: '%s' exists and is not a directory" % installdir
        sys.exit(1)

    #
    # Find all dynamic libraries in 'datadir' and look for libraries
    # there, also.
    #

    so_re = re.compile('.*\.so(\.\d+)*$')

    for root, dirs, files in os.walk(datadir):
        for f in files:
            # Skip the interpreter
            if f.startswith('ld-'):
                continue
            if so_re.match(f):
                fname = root + '/' + f
                install_libs(fname, installdir)

    # Install all the libs the binary itself needs
    install_libs(binary, installdir)

    # Install any /lib/libnss_* found
    # These match the glibc version we'll install.  This also might
    # install a libresolv and so forth when it searches dependencies
    dirs = os.listdir("/lib")
    for d in dirs:
        if d.startswith("libnss_") or d.startswith("libresolv"):
            d = "/lib/" + d
            print "Installing %s" % d
            shutil.copy2(d, installdir)
            # Recurse
            install_libs(d, installdir)

    # BINTAR created on Ubuntu 12.04 required /usr/lib/python2.x/config and
    # /usr/include/python2.x to the internal DATADIR of NovaAgent
    fix_ubuntu12_lib(datadir)
